/* MD5.C - RSA Data Security, Inc., MD5 message-digest algorithm
 */

/* Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
rights reserved.

License to copy and use this software is granted provided that it
is identified as the "RSA Data Security, Inc. MD5 Message-Digest
Algorithm" in all material mentioning or referencing this software
or this function.

License is also granted to make and use derivative works provided
that such works are identified as "derived from the RSA Data
Security, Inc. MD5 Message-Digest Algorithm" in all material
mentioning or referencing the derived work.

RSA Data Security, Inc. makes no representations concerning either
the merchantability of this software or the suitability of this
software for any particular purpose. It is provided "as is"
without express or implied warranty of any kind.

These notices must be retained in any copies of any part of this
documentation and/or software.
 */
#include <string.h>

#include <sys/heap.h>
#include <netinet/md5.h>

static void md5transform (unsigned long state[4], const unsigned char block[64]);
static inline void Encode (unsigned char *output, const unsigned long *input, unsigned int len);
static inline void Decode (unsigned long *output, const unsigned char *input, unsigned int len);

/* Constants for md5transform routine.
 */
static unsigned char Worder[64] = {
    0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
    1,6,11,0,5,10,15,4,9,14,3,8,13,2,7,12,
    5,8,11,14,1,4,7,10,13,0,3,6,9,12,15,2,
    0,7,14,5,12,3,10,1,8,15,6,13,4,11,2,9
};

static unsigned char Rorder[64] = {
    7,12,17,22, 7,12,17,22, 7,12,17,22, 7,12,17,22,
    5,9,14,20,  5,9,14,20,  5,9,14,20,  5,9,14,20,
    4,11,16,23, 4,11,16,23, 4,11,16,23, 4,11,16,23,
    6,10,15,21, 6,10,15,21, 6,10,15,21, 6,10,15,21
};

static unsigned long Korder[] = {
    0xd76aa478UL, 0xe8c7b756UL, 0x242070dbUL, 0xc1bdceeeUL,
    0xf57c0fafUL, 0x4787c62aUL, 0xa8304613UL, 0xfd469501UL,
    0x698098d8UL, 0x8b44f7afUL, 0xffff5bb1UL, 0x895cd7beUL,
    0x6b901122UL, 0xfd987193UL, 0xa679438eUL, 0x49b40821UL,
    0xf61e2562UL, 0xc040b340UL, 0x265e5a51UL, 0xe9b6c7aaUL,
    0xd62f105dUL, 0x02441453UL, 0xd8a1e681UL, 0xe7d3fbc8UL,
    0x21e1cde6UL, 0xc33707d6UL, 0xf4d50d87UL, 0x455a14edUL,
    0xa9e3e905UL, 0xfcefa3f8UL, 0x676f02d9UL, 0x8d2a4c8aUL,
    0xfffa3942UL, 0x8771f681UL, 0x6d9d6122UL, 0xfde5380cUL,
    0xa4beea44UL, 0x4bdecfa9UL, 0xf6bb4b60UL, 0xbebfbc70UL,
    0x289b7ec6UL, 0xeaa127faUL, 0xd4ef3085UL, 0x04881d05UL,
    0xd9d4d039UL, 0xe6db99e5UL, 0x1fa27cf8UL, 0xc4ac5665UL,
    0xf4292244UL, 0x432aff97UL, 0xab9423a7UL, 0xfc93a039UL,
    0x655b59c3UL, 0x8f0ccc92UL, 0xffeff47dUL, 0x85845dd1UL,
    0x6fa87e4fUL, 0xfe2ce6e0UL, 0xa3014314UL, 0x4e0811a1UL,
    0xf7537e82UL, 0xbd3af235UL, 0x2ad7d2bbUL, 0xeb86d391UL,
    0xe1f27f3aUL, 0xf5710fb0UL, 0xada0e5c4UL, 0x98e4c919UL
};

/* F, G, H and I are basic MD5 functions.
 */
#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z)))

/* ROTATE_LEFT rotates x left n bits.
 */
#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */
#define FF(a, b, c, d, x, s, ac) { \
    (a) += F ((b), (c), (d)) + (x) + (unsigned long)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}
#define GG(a, b, c, d, x, s, ac) { \
    (a) += G ((b), (c), (d)) + (x) + (unsigned long)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}
#define HH(a, b, c, d, x, s, ac) { \
    (a) += H ((b), (c), (d)) + (x) + (unsigned long)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}
#define II(a, b, c, d, x, s, ac) { \
    (a) += I ((b), (c), (d)) + (x) + (unsigned long)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}

/*!
 * \brief Convert 32bit value to byte array
 */
#ifndef __BIG_ENDIAN__
#define ulong2Array(aOutput, ulInput) { unsigned long __t = (ulInput); memcpy((aOutput), &__t, 4); }
#else
#define ulong2Array(aOutput, ulInput) { \
    (aOutput)[0] = (unsigned char)((unsigned char)(ulInput) & 0xff); \
    (aOutput)[1] = (unsigned char)(((unsigned long)(ulInput) >> 8) & 0xff); \
    (aOutput)[2] = (unsigned char)(((unsigned long)(ulInput) >> 16) & 0xff); \
    (aOutput)[3] = (unsigned char)(((unsigned long)(ulInput) >> 24) & 0xff); \
}
#endif

/*!
 * \brief Convert byte array to 32bit value
 */
#ifndef __BIG_ENDIAN__
#define Array2ulong(ulOutput, aInput) memcpy(&(ulOutput), aInput, 4);
#else
#define Array2ulong(ulOutput, aInput) { \
    (ulOutput) = ((unsigned long)((unsigned char)(aInput)[3] & 0xff)<<24) | \
                 ((unsigned long)((unsigned char)(aInput)[2] & 0xff)<<16) | \
                 ((unsigned long)((unsigned char)(aInput)[1] & 0xff)<<8)  | \
                 ((unsigned long)((unsigned char)(aInput)[0] & 0xff)); \
}
#endif

/* Encodes input (unsigned long) into output (unsigned char). Assumes len is
    a multiple of 4.
 */
static inline void Encode (unsigned char *output, const unsigned long *input, unsigned int len)
{
    unsigned int i, j;

    for (i = 0, j = 0; j < len; i++, j += 4) {
        ulong2Array(&output[j], input[i]);
    }
}

/* Decodes input (unsigned char) into output (unsigned long). Assumes len is
    a multiple of 4.
 */
static inline void Decode (unsigned long *output, const unsigned char *input, unsigned int len)
{
    unsigned int i, j;

    for (i = 0, j = 0; j < len; i++, j += 4) {
        Array2ulong(output[i], &input[j]);
    }
}

/* MD5 initialization. Begins an MD5 operation, writing a new context.
 */
void md5init(md5_context *context)
{
    context->count[0] = context->count[1] = 0;

    /* Load magic initialization constants.
     */
    context->state[0] = 0x67452301UL;
    context->state[1] = 0xefcdab89UL;
    context->state[2] = 0x98badcfeUL;
    context->state[3] = 0x10325476UL;
}

/* MD5 block update operation. Continues an MD5 message-digest
    operation, processing another message block, and updating the
    context.
 */
void md5update(md5_context *context, const unsigned char *input, const unsigned int inputLen)
{
    unsigned int i, index, partLen;

    /* Compute number of bytes mod 64 */
    index = (unsigned int)((context->count[0] >> 3) & 0x3F);

    /* Update number of bits */
    if ((context->count[0] += ((unsigned long)inputLen << 3)) < ((unsigned long)inputLen << 3))
        context->count[1]++;
    context->count[1] += ((unsigned long)inputLen >> 29);

    partLen = 64 - index;

    /* Transform as many times as possible.
     */
    if (inputLen >= partLen) {
        memcpy(&context->buffer[index], input, partLen);
        md5transform (context->state, context->buffer);

        for (i = partLen; i + 63 < inputLen; i += 64)
                md5transform (context->state, &input[i]);

        index = 0;
    }
    else {
        i = 0;
    }

    /* Buffer remaining input */
    memcpy((unsigned char *)&context->buffer[index], (unsigned char *)&input[i], inputLen-i);
}

/* MD5 finalization. Ends an MD5 message-digest operation, writing the
    the message digest and zeroizing the context.
 */
                                     /* message digest           context */
void md5final(unsigned char digest[16], md5_context *context)
{
    unsigned char bits[8];
    unsigned int index, padLen;

    /* Save number of bits */
    Encode (bits, context->count, 8);

    /* Pad out to 56 mod 64.
     */
    index = (unsigned int)((context->count[0] >> 3) & 0x3f);
    padLen = (index < 56) ? (56 - index) : (120 - index);

    unsigned char *pPadding = NutHeapAllocClear(64);
    pPadding[0] = 0x80;
    md5update (context, pPadding, padLen);
    NutHeapFree(pPadding);

    /* Append length (before padding) */
    md5update (context, bits, 8);
    /* Store state in digest */
    Encode (digest, context->state, 16);

    /* Zeroize sensitive information. */
    memset ((unsigned char *)context, 0, sizeof (*context));
}

/* MD5 basic transformation. Transforms state based on block.
 */
static void md5transform (unsigned long state[4], const unsigned char block[64])
{
    unsigned long temp;
    unsigned long a = state[0], b = state[1], c = state[2], d = state[3];
    unsigned long x[16];
    unsigned char i = 0;

    Decode (x, block, 64);

    for (; i < 16; ++i) {
        FF(a,b,c,d,x[Worder[i]],Rorder[i],Korder[i]);
        temp = d; d = c; c = b; b = a; a = temp;
    }

    for (; i < 32; ++i) {
        GG(a,b,c,d,x[Worder[i]],Rorder[i],Korder[i]);
        temp = d; d = c; c = b; b = a; a = temp;
    }

    for (; i < 48; ++i) {
        HH(a,b,c,d,x[Worder[i]],Rorder[i],Korder[i]);
        temp = d; d = c; c = b; b = a; a = temp;
    }

    for (; i < 64; ++i) {
        II(a,b,c,d,x[Worder[i]],Rorder[i],Korder[i]);
        temp = d; d = c; c = b; b = a; a = temp;
    }

    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;

    /* Zeroize sensitive information.
     */
    memset ((unsigned char *)x, 0, sizeof(x));
}

#ifdef MD5TEST
/*
 * Runs a selftest/verify
 *
 * If there is an error in the implementation this function returns -1
 * If all digests check out a 0 is returned.
 */
int md5test(void)
{
    static const struct {
        unsigned char *msg;
        unsigned char hash[16];
    } tTests[] = {
        { "",
        { 0xd4, 0x1d, 0x8c, 0xd9, 0x8f, 0x00, 0xb2, 0x04, 
                0xe9, 0x80, 0x09, 0x98, 0xec, 0xf8, 0x42, 0x7e } },
        { "a",
        {0x0c, 0xc1, 0x75, 0xb9, 0xc0, 0xf1, 0xb6, 0xa8, 
                0x31, 0xc3, 0x99, 0xe2, 0x69, 0x77, 0x26, 0x61 } },
        { "abc",
        { 0x90, 0x01, 0x50, 0x98, 0x3c, 0xd2, 0x4f, 0xb0, 
                0xd6, 0x96, 0x3f, 0x7d, 0x28, 0xe1, 0x7f, 0x72 } },
        { "message digest", 
        { 0xf9, 0x6b, 0x69, 0x7d, 0x7c, 0xb7, 0x93, 0x8d, 
                0x52, 0x5a, 0x2f, 0x31, 0xaa, 0xf1, 0x61, 0xd0 } }, 
        { "abcdefghijklmnopqrstuvwxyz",
        { 0xc3, 0xfc, 0xd3, 0xd7, 0x61, 0x92, 0xe4, 0x00, 
                0x7d, 0xfb, 0x49, 0x6c, 0xca, 0x67, 0xe1, 0x3b } },
        { "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
        { 0xd1, 0x74, 0xab, 0x98, 0xd2, 0x77, 0xd9, 0xf5, 
                0xa5, 0x61, 0x1c, 0x2c, 0x9f, 0x41, 0x9d, 0x9f } },
        { "12345678901234567890123456789012345678901234567890123456789012345678901234567890",
        { 0x57, 0xed, 0xf4, 0xa2, 0x2b, 0xe3, 0xc9, 0x55, 
                0xac, 0x49, 0xda, 0x2e, 0x21, 0x07, 0xb6, 0x7a } }, 
        { NULL, { 0 } }
    };

    unsigned short i;
    for (i = 0; tTests[i].msg != NULL; i++) {
        unsigned char aDigest[16];
        md5_context ctx;
    
        md5init( &ctx );
        md5update( &ctx, tTests[i].msg, strlen (tTests[i].msg) );
        md5final( aDigest, &ctx );

        if (memcmp(aDigest, tTests[i].hash, 16) != 0) {
            return (-1);
        }
    }
    return (0);
}
#endif /* MD5TEST */

